/*--[explode.c]----------------------------------------------------------------
 | Copyright (C) 2002 Dan A. Jackson
 |
 | This file is part of "clit" (Convert LIT). 
 |
 | "clit" is free software; you can redistribute it and/or modify
 | it under the terms of the GNU General Public License as published by
 | the Free Software Foundation; either version 2 of the License, or
 | (at your option) any later version.
 |
 | This program is distributed in the hope that it will be useful,
 | but WITHOUT ANY WARRANTY; without even the implied warranty of
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 | GNU General Public License for more details.
 |
 | You should have received a copy of the GNU General Public License
 | along with this program; if not, write to the Free Software
 | Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 | 
 | The GNU General Public License may also be available at the following
 | URL: http://www.gnu.org/licenses/gpl.html
*/
#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include "litlib.h"
#include "clit.h"

#ifdef _MSC_VER
#define strncasecmp _strnicmp
#endif


/*
 | Herein is the routines to handle "exploding" of .LIT files into their
 | component pieces.  
 | Some features, most noteably parsing the "/manifest" and expanding the
 | tag-compressed HTML are included in the library.
 */
manifest_type   manifest;
extern char *   writingFilename;

/* Leave off the newline intentionally */
const char * dtd_string = "<?xml version=\"1.0\"?>\n\
<!DOCTYPE package PUBLIC \n\
    \"+//ISBN 0-9673008-1-9//DTD OEB 1.0.1 Package//EN\"\n\
    \"http://openebook.org/dtds/oeb-1.0.1/oebpkg101.dtd\">";

int  dosify_mappings(manifest_type * pManifest);
char * strmerge(char * head, char * body, char * tail);
char * get_opf_name(char * pathOutput, char * litname);
int  write_htmlish_file(lit_file * litfile, char *, char *, int);
int  write_raw_file(lit_file * litfile,char *pathInternal, char *pathExternal);

int explode_lit(lit_file * litfile, char * litName, char * pathOutput) 
{
    char            * pathOPF = NULL;
    char            * pathInternal = NULL;
    char            * path = NULL;
    int             status;
    int             stage, i;

    status = lit_read_manifest(litfile, &manifest);
    if (status) goto bad;

    status = dosify_mappings(&manifest);
    if (status) goto bad;

    pathOPF = get_opf_name(pathOutput,litName); 
    writingFilename = pathOPF;
    status = write_htmlish_file(litfile, "/meta", pathOPF, 1);
    free(pathOPF); pathOPF = NULL;
    writingFilename = NULL; 
    if (status) goto bad;

    for (stage = 0; stage < NUM_MANIFEST_MAPS; stage++) {
        name_mapping * map;

        if (!manifest.mappings[stage]) continue;
        map = manifest.mappings[stage];
        for (i = 0; i < manifest.num_mappings[stage]; i++) {
            path = strmerge(pathOutput,(char *)map->sOriginal,NULL);
            if (!path) {
                status = E_LIT_OUT_OF_MEMORY;
                goto bad;
            }
            writingFilename = path;
            show_information("Writing out \"%s\" as \"%s\" ...",
                map->sInternal, map->sOriginal);

            status = E_LIT_INTERNAL_ERROR;
            switch (stage) {
                case 0:
                case 1:
                    pathInternal = strmerge("/data/",(char *)map->sInternal,
                        "/content");
                    if (!pathInternal) { 
                        status = E_LIT_OUT_OF_MEMORY;
                        goto bad;
                    }
                    status = write_htmlish_file(litfile,pathInternal,path,0);
                    break; 
                case 2:
                case 3:
                    pathInternal = strmerge("/data/",(char *)map->sInternal,0);
                    if (!pathInternal) {
                        status = E_LIT_OUT_OF_MEMORY;
                        goto bad;
                    }
                    status = write_raw_file(litfile,pathInternal,path);
                    break;
            }
            if (status) {
                if (status > 0) {
                     show_information("Wrote out \"%s\", but with errors.\n",
                         path);
                }
                if (status < 0) return -1;
            }
            else {
                show_information("Successfully written to \"%s\".\n", path);
            }
            if (path) { free(path); path = NULL;}
            if (pathInternal) { free(pathInternal); pathInternal = NULL; }
            writingFilename = NULL;
            map++;
 
        }
    }
    
    return  0;  
bad:
    if (path)         free(path);
    if (pathOPF)      free(pathOPF);
    if (pathInternal) free(pathInternal);
    return status;
}

/*
 | FUTUREFIX - Should this routine filter out UTF8?
 | FUTUREFIX - Should this routine restore %20, %2E etc in names?  */
int dosify_mappings(manifest_type * pManifest)
{
    int    i, j, len, idx, pos;
    U8     * name, b;

    for (j = 0; j < NUM_MANIFEST_MAPS; j++)
    {
        name_mapping * map;

        if (!pManifest->mappings[j]) continue;
        map = pManifest->mappings[j];
        for (i = 0; i < pManifest->num_mappings[j]; i++)
        {
            name = map->sOriginal;
            idx = len = strlen((char *)name);
            pos = 0;
            while (idx) {
                if ((name[idx] == '/') ||
                    (name[idx] == '\\') ||
                    (name[idx] == ':')) {
                    idx++;
                    break; 
                }
                idx--;
            } 
            for (pos = 0; idx < len; idx++) {
                b = name[idx];
                if ((b <  ' ')||
                    (b == '?')||
                    (b == '<')||
                    (b == '>')||
                    (b =='\"')||
                    (b == '*')||
                    (b == '|')||
                    (b == ';')||
                    (b ==0x7f)) b = '_';
                name[pos++] = b;
            }
            name[pos] = '\0';
            map++;
        }
    }
    return 0; 
}


char * get_opf_name(char * pathOutput, char * litname)
{
    int idx, len;
    char * str;

    idx = len = strlen(litname);

    /* assume that litname is a valid LIT file */
    while (idx) {
        if ((litname[idx] == '/') ||
            (litname[idx] == '\\') || 
            (litname[idx] == ':'))  {
            idx++;
            break;
        }
        idx--;
    }
    if (strncasecmp(&litname[len-4],".lit",4) == 0) {
        str = strmerge(pathOutput,&litname[idx],NULL);
        if (!str) return str;
        strcpy(&str[strlen(str)-3], "opf");
        return str;
    } 
    return strmerge(pathOutput,&litname[idx],".opf");
}

char * strmerge(char * head, char * body, char * tail)
{
    int len;
    char * str;

    len = 0;
    if (head) len += strlen(head);
    if (body) len += strlen(body);
    if (tail) len += strlen(tail); 
    len++;

    str = malloc(len);
    if (!str) {
        lit_error(0,"Unable to malloc %d bytes for a string.",len);
        return NULL;
    }
    str[0] = '\0';
    if (head) strcat(str,head);
    if (body) strcat(str,body);
    if (tail) strcat(str,tail);
    return str;
}

int write_callback(void * v, U8 * pData, int nBytes)
{
    FILE    * f;
    int     r;

    if (!v) { 
        f = stdout; 
    } else {
        f = (FILE *)v;
    }
    r = fwrite(pData,1, nBytes,f);
    if (r < 0) {
        lit_error(ERR_W|ERR_LIBC,"fwrite of %d bytes failed!",nBytes);
    }
    return r;
}


int  write_htmlish_file(lit_file * litfile, char * pathInternal,
    char * pathExternal, int type)
{
    U8  * p;
    int nbytes;
    int status;
    FILE * fOut;

    status = lit_get_file(litfile,pathInternal,&p, &nbytes);
    if (status) return status;

    fOut = fopen(pathExternal, "w");
    if (!fOut) {
        free(p);
        lit_error(ERR_LIBC|ERR_W,"fopen(%s) failed!", pathExternal);
        return -1;
    }
    if (type) {
        status = write_callback((void *)fOut,(U8 *)dtd_string, 
            strlen(dtd_string));
        if ((size_t)status != strlen(dtd_string)) {free(p);return -1;    }
    }

    status = lit_reconstitute_html(p, 0, nbytes, type, &manifest, 
        write_callback, (void *)fOut);

    if ((status > 0) && (status != nbytes)) {
        lit_error(ERR_R,
"Warning - Couldn't fully expand \"%s\", decoded %d out of %d.\n",
            pathInternal, status, (int)nbytes);
        status = 1;
    }
    else if (status < 0) {
        lit_error(ERR_R,
"ERROR - Failure during tag expansion of \"%s\" Code: %d!\n", pathInternal, 
            status);
    }
    else {
        status = 0;
    }
    fclose(fOut);
    free(p);
    return status;
}

int write_raw_file(lit_file * litfile,char * pathInternal, char * pathExternal)
{
    int     status;
    int     nbytes;
    U8      * p;
    FILE * fOut;

    status = lit_get_file(litfile,pathInternal,&p,&nbytes);
    if (status) return status;

    fOut = fopen(pathExternal, "wb");
    if (!fOut) {
        free(p);
        lit_error(ERR_LIBC|ERR_W,"fopen(%s) failed!", pathExternal);
        return -1;
    }
    status = fwrite(p,1, nbytes,fOut);
    free(p); p = NULL;

    if (status < 0) {
        lit_error(ERR_W|ERR_LIBC,"Failed to write %d bytes.",nbytes);
        return status;
    }
    if (status != nbytes) {
        lit_error(ERR_W,"Partial write, wrote %d out of %d.",status,nbytes);
        return -1;
    }
    fclose(fOut);
    return 0;
}


